;; This implementation of ART Processor is capable of 
;; retrieving body data. It can also interpolate missing body samples.
;; @field ID The id of the ARTbody to retrieve.
;;
PRO ART_BODY_PROCESSOR__DEFINE

    ; define the class structure 
    struct = { ART_BODY_PROCESSOR, INHERITS art_processor, id:0 }

END


;; Creates a new ART Body Processor. 
;; The ART Body Processor is optionally bound to the ART Camera Data given. 
;; The body id can also be predefined optionally.
;; @param DATA (?IN) The ART Camera Data to setup initially. (See superclass.)
;; @key ID (?IN) The body id to setup initially. Defaults to 0. (See method set_body_id.)
;;
FUNCTION ART_BODY_PROCESSOR::INIT, DATA, ID=ID

    ; initialize superclass
    if (self->art_processor::init(data) eq 0) then begin
        return, 0
    endif 

    ; try to set initial body id
    catch, e
    if (e ne 0) then begin

        message, !error_state.msg, /informational
        return, 0

    endif

        ; try to set initial body id
        self->set_body_id, id

    catch, /cancel
    return, 1

END


;; Cleans up upon destruction.
;; This ART Body Processor does not need to perform any clean up itself.
;; However, the cleanup method of the superclass is invoked.
;;
PRO ART_BODY_PROCESSOR::CLEANUP

    self->art_processor::cleanup

END


;; Sets up the body id of the ART body to retrieve.
;; This value is passed to the get_body method of the ART Camera Data upon processing.
;; If no value is given, the body id remains unchanged.
;; TODO Check that the given id is a positive integer number.
;; @param ID (?IN) The body id to setup this ART Body Processor for.
;;                 This must be a positive integer number.
;;
PRO ART_BODY_PROCESSOR::SET_BODY_ID, ID

    ; is any parameter given?
    if (n_elements(id) eq 0) then begin
        return
    endif

    self.id = fix(id)

END


;; Returns the body id setup for this ART Body Processor. 
;; @return The body id.
;;
FUNCTION ART_BODY_PROCESSOR::GET_BODY_ID

    return, self.id

END


;; A processing method for retrieving a sequence of ART body locations and orientations.
;; This method first gets the body data for the setup id from the setup ART Camera Data. 
;; It then interpolates missing samples for that body, and returns the result.
;; @return A three-dimensional array of body data. 
;;         The first dimension corresponds to the sample frames. 
;;         The second and third dimensions correspond to a 4x4 homogenous matrix describing
;;         the location and orientation of the body.
;;
FUNCTION ART_BODY_PROCESSOR::RETRIEVE_BODY

    ; is this processor setup properly?
    if (self->is_processible() eq 0) then begin
        message, 'Setup all fields of this ART_BODY_PROCESSOR before using it!'
    endif
    
    ; use the method of the superclass to get raw body
    body = (self->get_art_data())->get_body(self.id)

    ; interpolate any missing samples of the body
    self->interpolate_missing_bodies, body

    ; return interpolated body
    return, body

END


;; An internal method interpolating missing samples in an array of ART body locations and orientations.
;; Each body sample containing NaNs is interpolated using valid data from nearby sample frames. 
;; The interpolation is done by nearest neighbor sampling. Note that this might lead to significant
;; jumps in the middle of long ranges with missing data.
;; TODO Implement linear interpolation? (Orthogonality of the matrix must be preserved!)
;; @param BODIES (IN/OUT) A three-dimensional array of body data, as returned by the get_body method of ART Camera Data.
;;
PRO ART_BODY_PROCESSOR::INTERPOLATE_MISSING_BODIES, BODIES

    ; check body array 
    if (size(bodies, /n_dimensions) ne 3) then begin
        message, 'The array of bodies must have exactly three dimensions! (Sample frames, homogenous matrix rows, homogenous matrix columns)'
    endif
    body_dimensions = size(bodies, /dimensions)
    if ((body_dimensions[1] ne 4) or (body_dimensions[2] ne 4)) then begin
        message, 'The array of bodies must contain 4x4 homogenous matrices!'
    endif

    ; find any missing body samples
    n_frames = body_dimensions[0]
    missing = where(finite(total(total(bodies, 3), 2), /nan), n_missing)
    
    ; do any body samples have to be interpolated?
    if (n_missing gt 0) then begin
    
        ; get frame numbers adjunctant to frames where the body is missing
        lower_margin_indices = where(missing ne [-1, missing+1])
        lower_margins = missing[lower_margin_indices] - 1
        upper_margin_indices = where(missing ne [missing[1:*]-1, -1])
        upper_margins = missing[upper_margin_indices] + 1

        ; assert that everything is still ok
        if (n_elements(lower_margins) ne n_elements(upper_margins)) then begin
            message, 'Internal error: Calculated numbers of lower and upper margins are not equal!'
        endif

        ; iterate over each frame with a missing body sample
        lower_margin_index = 0
        upper_margin_index = 0
        for frame_index = 0ul, n_elements(missing)-1 do begin

            ; get numer of current frame, and numbers of next frames with valid bodies 
            frame = missing[frame_index]
            if (lower_margin_index+1 lt n_elements(lower_margins)) then begin
                if (lower_margins[lower_margin_index+1] le frame) then begin
                    lower_margin_index += 1 
                endif
            endif
            lower_margin = lower_margins[lower_margin_index]
            if (upper_margins[upper_margin_index] le frame) then begin
                upper_margin_index += 1 
            endif
            upper_margin = upper_margins[upper_margin_index]

            ; is there any data at all to sample from?
            if ((lower_margin lt 0) and (upper_margin ge n_frames)) then begin
                ; reset missing body to the identity
                bodies[*,*,*] = rebin(reform([[1.0,0.0,0.0,0.0], [0.0,1.0,0.0,0.0], [0.0,0.0,1.0,0.0], [0.0,0.0,0.0,1.0]], 1, 4, 4), n_frames, 4, 4)
                message, 'No valid data to interpolate body from. Resetting it to identity matrices!' , /informational
                break
            endif
            
            ; sample from left?
            if (((lower_margin ge 0) and ((frame - lower_margin) le (upper_margin - frame))) or (upper_margin ge n_frames)) then begin
                bodies[frame,*,*] = bodies[lower_margin,*,*]
                continue
            endif 
            
            ; sample from right
            bodies[frame,*,*] = bodies[upper_margin,*,*]

        endfor   
    
    endif

END
