;; This implementation of ART Processor is capable of 
;; retrieving marker data and applying coordinate system transformations to them. 
;; As its superclass, it can interpolate missing marker samples and can be setup 
;; to filter for relevant markers.
;; @field BODY_PROCESSOR A reference to an internal ART Body Processor, used to retrieve reference targets.
;; @field CALIBRATION_MATRIX A 4x4 homogenous calibration matrix describing the transformation to the target coordinate system.
;;
PRO ART_CALIBRATING_MARKER_PROCESSOR__DEFINE

    ; define the class structure 
    struct = { ART_CALIBRATING_MARKER_PROCESSOR, INHERITS art_marker_processor, body_processor:obj_new(), calibration_matrix:[[1.0,0.0,0.0,0.0],[0.0,1.0,0.0,0.0],[0.0,0.0,1.0,0.0],[0.0,0.0,0.0,1.0]] }

END


;; Creates a new ART Calibrating Marker Processor. 
;; The ART Calibrating Marker Processor is optionally bound to the ART Camera Data given. 
;; The distance_threshold , the list of relevant markers, the refernce body id and the calibration matrix 
;; can also be predefined optionally.
;; This constructor also creates a new ART Body Processor bound to the same ART Camera DATA, which is later 
;; used internally retrieve reference targets.
;; @param DATA (?IN) The ART Camera Data to setup initially. (See method set_art_data.)
;; @key DISTANCE_THRESHOLD (?IN) The distance_threshold to setup initially. (See superclass.)
;; @key RELEVANT_MARKERS (?IN) The list of relevant markers to setup initially. (See superclass.)
;; @key REFERENCE_BODY_ID (?IN) The reference body id to setup initially. (See method set_reference_body_id.)
;; @key CALIBRATION_MATRIX (?IN) The calibration matrix to setup initially. (See method set_calibration_matrix.)
;;
FUNCTION ART_CALIBRATING_MARKER_PROCESSOR::INIT, DATA, DISTANCE_THRESHOLD=DISTANCE_THRESHOLD, RELEVANT_MARKERS=RELEVANT_MARKERS, REFERENCE_BODY_ID=REFERENCE_BODY_ID, CALIBRATION_MATRIX=CALIBRATION_MATRIX

    ; initialize superclass
    if (self->art_marker_processor::init(data, distance_threshold=distance_threshold, relevant_markers=relevant_markers) eq 0) then begin
        return, 0
    endif 

    ; initialize adapted body processor
    self.body_processor = obj_new('art_body_processor', data, id=refernce_body_id)
    if (obj_valid(self.body_processor) eq 0) then begin
        return, 0
    endif 

    ; try to initialize field values
    catch, e
    if (e ne 0) then begin

        message, !error_state.msg, /informational
        return, 0

    endif

        ; try to initialize field values
        self->set_calibration_matrix, calibration_matrix

    catch, /cancel
    return, 1

END


;; Cleans up upon destruction.
;; More specifically, the ART Body Processor used internally to retrieve 
;; refernce targets is destroyed. 
;; Also, the cleanup method of the superclass is invoked.
;;
PRO ART_CALIBRATING_MARKER_PROCESSOR::CLEANUP

    self->art_marker_processor::cleanup
    obj_destroy, self.body_processor

END


;; Sets up the ART Camera Data to be processed by this ART Calibrating Marker Processor. 
;; Fistly the corresponding method of the superclass is invoked. Secondly the corresponding
;; method on the internally used ART Body Processor is invoked, to make sure that it always 
;; uses the the same ART Camera Data as this ART Calibrating Marker Processor.
;; @param DATA (?IN) The ART Camera Data to setup this ART Calibrating Marker Processor for.
;;
PRO ART_CALIBRATING_MARKER_PROCESSOR::SET_ART_DATA, DATA

    self->art_marker_processor::set_art_data, data
    if (obj_valid(self.body_processor)) then begin
        self.body_processor->set_art_data, data
    endif

END


;; Sets up the body id of the ART body to use as a refernce target.
;; This value is passed on to the internally used ART Body Processor.
;; @param ID (?IN) The body id of the refernce target.
;;
PRO ART_CALIBRATING_MARKER_PROCESSOR::SET_REFERENCE_BODY_ID, ID

    self.body_processor->set_body_id, id

END


;; Returns the body id of the refernce target setup for this ART Calibrating Marker Processor. 
;; @return The body id of thr refernce target.
;;
FUNCTION ART_CALIBRATING_MARKER_PROCESSOR::GET_REFERENCE_BODY_ID

    return, self.body_processor->get_body_id()

END


;; Sets up the calibration matrix describing the transformation to the target coordinate system.
;; If no value is given, the calibration matrix is reset to a 4x4 homogenous identity matrix.
;; TODO Check that a homogenous matrix in form of a 4x4 numeric array is given.
;; @param CALIBRATION_MATRIX (?IN) The calibration matrix to setup this ART Calibrating Marker Processor for.
;;                                 This must be a numeric 4x4 array describing a homogenous matrix.
;;
PRO ART_CALIBRATING_MARKER_PROCESSOR::SET_CALIBRATION_MATRIX, CALIBRATION_MATRIX

    ; reset to identity, if no value given
    if (n_elements(calibration_matrix) eq 0) then begin
        self.calibration_matrix = [[1,0,0,0],[0,1,0,0],[0,0,1,0],[0,0,0,1]]
        return
    endif

    self.calibration_matrix = calibration_matrix

END


;; TODO Implement calibration matrix chaining.
;;
PRO ART_CALIBRATING_MARKER_PROCESSOR::APPEND_CALIBRATION_MATRIX, CALIBRATION_MATRIX

    message, 'This method is not implemented yet. Doing nothing!', /informational

END


;; Returns the calibration matrix setup for this ART Calibrating Marker Processor. 
;; @return The calibration matrix.
;;
FUNCTION ART_CALIBRATING_MARKER_PROCESSOR::GET_CALIBRATION_MATRIX

    return, self.calibration_matrix

END


;; Loads the calibration matrix from a SAV file.
;; The file must contain data in IDL's SAV format, in a certain structure. The required structure is the same 
;; as created by the save_calibration_matrix method. 
;; TODO Improve checks for integrity of loaded data.
;; @param FILENAME (IN) The full path to the SAV file. It can be absolute or relative.
;;
PRO ART_CALIBRATING_MARKER_PROCESSOR::LOAD_CALIBRATION_MATRIX, FILENAME

    ; show file picker dialog if needed
    if (n_elements(filename) eq 0) then begin
        filename = dialog_pickfile(title='Load Calibration Matrix From File', filter=['*.sav'], /must_exist, /fix_filter)
    endif

    ; check file
    if (filename eq '') then begin
        message, 'No valid file was specified!'
    end

    ; check file format by filename extension
    if (strmatch(filename, "?*.sav", /fold_case) eq 0) then begin
        message, 'Given file does not have the recommended suffix ".SAV". Trying to read calibration matrix anyway...', /informational
    endif

    ; restore calibration matrix from file 
    restore, filename
    
    ; check calibration matrix
    if (n_elements(calibration_matrix) eq 0) then begin
        message, 'Given file does not contain a valid calibration matrix!'
    endif

    ; set the new calibration matrix
    self->set_calibration_matrix, calibration_matrix

END


;; Saves the calibration matrix to a SAV file.
;; The created file uses IDL's SAV format, in a certain structure. The contents of the file can be restored 
;; by the load_calibration_matrix method later.
;; @param FILENAME (IN) The full path to the SAV file. It can be absolute or relative.
;;
PRO ART_CALIBRATING_MARKER_PROCESSOR::SAVE_CALIBRATION_MATRIX, FILENAME

    ; show file picker dialog if needed
    if (n_elements(filename) eq 0) then begin
        filename = dialog_pickfile(title='Save Calibration Matrix To File', filter=['*.sav'], /fix_filter)
    endif

    ; check file
    if (filename eq '') then begin
        message, 'No valid file was specified!'
    end

    ; check file format by filename extension
    if (strmatch(filename, "?*.sav", /fold_case)) then begin
        message, 'Given file does not have the recommended suffix ".SAV". Saving calibration matrix anyway...', /informational
    endif

    ; get calibration matrix
    calibration_matrix = self->get_calibration_matrix()

    ; save calibration matrix to file
    save, calibration_matrix, filename=filename

END


;; A processing method for retrieving a sequence of ART marker locations 
;; transformed to a well-defined coordinate system.
;; This method first retrieves, filters and interpolates marker data, the same way as its 
;; superclass does. Second it uses the internal ART Body Processor to retrieve the data of the 
;; setup refernce body. 
;; Then for each sample frame, the marker data are transformed from the global 
;; ART coordinate system to the local coordinate system of the refernce body. (Note that this 
;; needs to be done per sample, since the refernce body is not necesserilly fixed with respect to the
;; ART coordinate system.) If no valid refernce body is setup, the marker data are not transformed in this step.
;; Finally the setup calibration matrix is applied to each ART marker sample. This transforms the marker data 
;; from the coordinate system of the refernce body, to the desired target coordinate system. 
;; @return A three-dimensional array of ART marker data with the above coordinate system transformations applied.
;;         The first dimension corresponds to the sample frames. 
;;         The second dimension corresponds to the slots of relevant markers. 
;;         The third dimension corresponds to the 3D spatial coordinates of the markers.
;;
FUNCTION ART_CALIBRATING_MARKER_PROCESSOR::RETRIEVE_MARKERS

    ; is this processor setup properly?
    if (self->is_processible() eq 0) then begin
        message, 'Setup all fields of this ART_CALIBRATING_MARKER_PROCESSOR before using it!'
    endif
    
    ; use the method of the superclass to get markers
    markers = self->art_marker_processor::retrieve_markers()

    ; apply coordinate system transformations from refernce body
    self->apply_coordinate_system, markers, self.body_processor->retrieve_body()

    ; apply coordinate system transformations from calibration matrix
    self->apply_coordinate_system, markers, reform(self->get_calibration_matrix(), 1, 4, 4)

    ; return processed markers
    return, markers
    
END


;; This internal method applies a sequence of ART body coordinate system transformations to a sequence of ART markers.
;; Each of the given body transformations is applied to all of the markers in the corresponding sample frame. If the number of bodies
;; does not correspond to the number sample frames implied by the markers, it is adjusted by interpolation. As a special case, 
;; if the sequence of bodies contains only one entry, the corresponding transformation is applied to all markers. (Note that this 
;; method will fail, if any of the bodies or markers contain NaN data.)
;; TODO Interpolate transformations more smartly! Orthogonality should be preserved!
;; @param MARKERS (IN/OUT) A three-dimensional array of ART marker data, as returned by the get_markers method of ART Camera Data.
;; @param BODIES (IN) A three-dimensional array of body data, as returned by the get_body method of ART Camera Data.
;;
PRO ART_CALIBRATING_MARKER_PROCESSOR::APPLY_COORDINATE_SYSTEM, MARKERS, BODIES

    ; check marker array
    if (size(markers, /n_dimensions) ne 3) then begin
        message, 'The array of markers must have exactly three dimensions! (Sample frames, markers, 3D-coordinates)'
    endif
    marker_dimensions = size(markers, /dimensions)
    if (marker_dimensions[2] ne 3) then begin
        message, 'The array of markers must contain 3d vectors!'
    endif

    ; check body array 
    if (size(bodies, /n_dimensions) ne 3) then begin
        message, 'The array of bodies must have exactly three dimensions! (Sample frames, homogenous matrix rows, homogenous matrix columns)'
    endif
    body_dimensions = size(bodies, /dimensions)
    if ((body_dimensions[1] ne 4) or (body_dimensions[2] ne 4)) then begin
        message, 'The array of bodies must contain 4x4 homogenous matrices!'
    endif

    ; resample the array of transformations to have the same amount of frames as the array of markers
    n_frames = marker_dimensions[0]
    transformations = congrid(bodies, n_frames, 4, 4, /center)

    ; iterate over all samples
    for i = 0ul, n_frames-1 do begin

        ; get transformation matrix for current sample
        trans = transpose(transformations[i,*,*], [1,2,0])
        trans = invert(trans)
        
        ; iterate over each marker of current sample
        for j = 0ul, marker_dimensions[1]-1 do begin
            marker = transpose([transpose(markers[i,j,*]), 1.0])
            marker = trans ## marker
            markers[i,j,*] = marker[0:2] / marker[3]
        endfor
    endfor

END


;; Checks if this ART Calibrating Marker Processor is setup correctly. 
;; Firstly the corresponding method of the superclass is invoked. Secondly the
;; corresponding method of the internally used ART Body Processor is invoked.
;; The result of this method is a boolean AND of those two values.
;; @return 1, if this ART Calibrating Marker Processor is setup correctly, 0 otherwise.
;;
FUNCTION ART_CALIBRATING_MARKER_PROCESSOR::IS_PROCESSIBLE

    ; are both the superclass and the adapted body processor processible?
    return, self->art_marker_processor::is_processible() and self.body_processor->is_processible()

END
