;; This class represents a sequence of 3-dimensional ART Camera Data.
;; Objects of this class store the ART Camera Data internally using a similar structure as the ART ASCII format. 
;; The data are basically organised in sequential sample frames. For each of them a timestamp, the locations of 
;; ART markers, and the locations and orientations of ART bodies are stored. Each sample can contain a different amount 
;; of markers and bodies. Other data, like measurement tools, flysticks, gloves, or calibration bodies are currently not supported.
;; Objects of this class are immutable. That is, the ART Camera Data they represent cannot be modified once created. However, several
;; data access methods are offered allowing retrieve the ART Camera Data in different ways.
;; @field DATA Pointer to contain a sequence of ART Camera Data samples. 
;;             Each sample is represented as a structure of the type ART_CAMERA_SAMPLE. Such structures may in turn contain
;;             arrays of ART_CAMERA_MARKERs and ART_CAMERA_BODYs, representing the ART markers and ART bodies present in the
;;             given sample frame.
;;
PRO ART_CAMERA_DATA__DEFINE

    ; define auxilliary structures
    struct = {art_camera_marker, mar:0L, qu:0.0, loc:[0.0, 0.0, 0.0]}
    struct = {art_camera_body, bod:0L, qu:0.0, loc:[0.0, 0.0, 0.0], ang:[0.0, 0.0, 0.0], rot:[[0.0, 0.0, 0.0], [0.0, 0.0, 0.0], [0.0, 0.0, 0.0]]}
    struct = {art_camera_sample, frame:0L, ts:0.0, nbodcal:0, nbod:0, nfly:0, nmea:0, nmar:0, ngl:0, markers:ptr_new(), bodies:ptr_new()}


    ; define the class structure 
    struct = { ART_CAMERA_DATA, data:ptr_new() }

END


;; Creates new ART Camera Data according to the contents of a given file.
;; This constructor calls the load method to read the ART Camera Data, and hence 
;; accepts the same arguments. The contents of the file are stored internally, and can later 
;; be retrieved by several data access methods. Once the the ART Camera Data are loaded, this 
;; object cannot be modified, since it is immutable.
;; @param FILENAME (?IN) See load method.
;; @key ASCII (?IN) See load method.
;; @key SAV (?IN) See load method.
;;
FUNCTION ART_CAMERA_DATA::INIT, FILENAME, ASCII=ASCII, SAV=SAV

    ; try to read ART camera data from a file
    catch, e
    if (e ne 0) then begin

        message, !error_state.msg, /informational
        return, 0

    endif

        ; try to read ART camera data from a file
        self->load, filename, ascii=ascii, sav=sav
        return, 1

    catch, /cancel

END


;; Cleans up the heap upon destruction.
;; More specifically all pointers used internally to represent the 
;; ART Camera Data are freed recursively.
;;
PRO ART_CAMERA_DATA::CLEANUP

    if ptr_valid(self.data) then begin
        heap_free, self.data
    endif

END


;; Loads ART Camera Data from a file, supporting different file formats.
;; According the filename suffix or to the specified format keys data are
;; loaded using specific file handlers. For ART ASCII files (*.art, *.txt) the
;; method load_art_ascii is called. For ART SAV files (*.sav) the method
;; load_art_sav is called. 
;; @param FILENAME (?IN) The full path to the ASCII file. It can be absolute or relative.
;;                       If the filename is ommitted, a file picker dialog is displayed,
;;                       so the user can specify the file to load from.
;; @key ASCII (?IN) Set this format flag to force the file to be considered in ART ASCII format.
;; @key SAV (?IN) Set this format flag to force the file to be considered in SAV format.
;;
PRO ART_CAMERA_DATA::LOAD, FILENAME, ASCII=ASCII, SAV=SAV

    ; check if ART data have already been loaded
    if (ptr_valid(self.data)) then begin
        message, 'Cannot load ART data from file, since some data are already loaded!'
    endif

    ; show file picker dialog if needed
    if (n_elements(filename) eq 0) then begin
        filename = dialog_pickfile(title='Load ART Camera Data From File', filter=['*.art;*.txt', '*.sav'], /read, /must_exist, /fix_filter)
    endif

    ; check file
    if (filename eq '') then begin
        message, 'No valid file was specified!'
    end

    ; determine file format by filename extension
    if (strmatch(filename, "?*.art", /fold_case) or strmatch(filename, "?*.txt", /fold_case)) then begin
        format = 'ASCII'
    endif else if (strmatch(filename, "?*.sav", /fold_case)) then begin
        format = 'SAV'
    endif else begin
        format = ''
    endelse

    ; override file format by given flags
    if (keyword_set(ascii)) then begin
        format = 'ASCII'
    endif else if (keyword_set(sav)) then begin
        format = 'SAV'
    endif

    ;load file according to specified format
    if (format eq 'ASCII') then begin
        self->load_ascii, filename
    endif else if (format eq 'SAV') then begin
        self->load_sav, filename
    endif else begin
        message, 'Could not determine file format!'
    endelse 

END


;; Loads ART camera data from an ASCII file.
;; The file must contain data in the ART ASCII format as specified in appendix B.1 of the ARTtrack & DTrack Manual (v1.22).
;; TODO Speed up this function by use of more efficient array operations.
;; TODO add checks for correct file format and issue a warning if no data could be loaded.
;; @param FILENAME (IN) The full path to the ASCII file. It can be absolute or relative.
;;
PRO ART_CAMERA_DATA::LOAD_ASCII, FILENAME

    ; check if ART data have already been loaded
    if (ptr_valid(self.data)) then begin
        message, 'Cannot load ART data from file, since some data are already loaded!'
    endif

    ;profiling code
    t0 = systime(1)

    ; define data structures used
    blank_pattern = '^ *$'
    acm_pattern = 'mar *([^ ]*) *qu *([^ ]*) *loc *([^ ]*) *([^ ]*) *([^ ]*)'
    acb_pattern = 'bod *([^ ]*) *qu *([^ ]*) *loc *([^ ]*) *([^ ]*) *([^ ]*) *ang *([^ ]*) *([^ ]*) *([^ ]*) *rot *([^ ]*) *([^ ]*) *([^ ]*) *([^ ]*) *([^ ]*) *([^ ]*) *([^ ]*) *([^ ]*) *([^ ]*)'
    acs_pattern = 'frame *([^ ]*) *ts *([^ ]*) *nbodcal *([^ ]*) *nbod *([^ ]*) *nfly *([^ ]*) *nmea *([^ ]*) *nmar *([^ ]*) *ngl *([^ ]*)'
    blank_pattern = '^ *$'
    art_camera_data = -1
    art_camera_sample = -1
    art_camera_markers = -1
    art_camera_bodies = -1

    ; read data
    openr, lun, filename, /get_lun
    while (eof(lun) ne 1) do begin
        
        ; read data from file line-wise
        current_line = ''
        readf, lun, current_line, format='(a)' 

        ; was a blank line encountered?
        blank_line = stregex(current_line, blank_pattern, /boolean)
        if (blank_line) then begin
            ;message, 'A blank line was encountered.', /informational 

            ; finish the preceeding art camera sample, if present
            if (n_tags(art_camera_sample) gt 0) then begin
                ;message, 'Finishing an ART camera sample...', /informational
                ; are there any markers to add do the camera data sample?
                if (n_tags(art_camera_markers) eq 0) then begin
                    if (art_camera_sample.nmar ne 0) then begin
                        message, 'No markers could be read for sample frame ' + int2str(art_camera_sample.frame) + ', though the header indicates there are ' + int2str(art_camera_sample.nmar), /informational
                    endif
                endif else begin
                    art_camera_sample.markers = ptr_new(art_camera_markers)
                    if (art_camera_sample.nmar ne n_elements(art_camera_markers)) then begin
                        message, int2str(n_elements(art_camera_markers)) + ' markers were read for sample frame ' + int2str(art_camera_sample.frame) + ', though the header indicates there are ' + int2str(art_camera_sample.nmar), /informational
                    endif
                endelse
                ; are there any bodies to add to the data sample?
                if (n_tags(art_camera_bodies) eq 0) then begin
                    if (art_camera_sample.nbod ne 0) then begin
                        message, 'No bodies could be read for sample frame ' + int2str(art_camera_sample.frame) + ', though the header indicates there are ' + int2str(art_camera_sample.nbod), /informational
                    endif
                endif else begin
                    art_camera_sample.bodies = ptr_new(art_camera_bodies)
                    if (art_camera_sample.nbod ne n_elements(art_camera_bodies)) then begin
                        message, int2str(n_elements(art_camera_bodies)) + ' bodies were read for sample frame ' + int2str(art_camera_sample.frame) + ', though the header indicates there are ' + int2str(art_camera_sample.nbod), /informational
                    endif
                endelse
                ; are there any measurement tools?
                if (art_camera_sample.nmea gt 0) then begin
                    message, 'Ignored any measurement tools for sample frame ' + int2str(art_camera_sample.frame) + ', though the header indicates that there are ' + int2str(art_camera_sample.nmea), /informational
                endif
                ; are there any flysticks?
                if (art_camera_sample.nfly gt 0) then begin
                    message, 'Ignored any flysticks for sample frame ' + int2str(art_camera_sample.frame) + ', though the header indicates that there are ' + int2str(art_camera_sample.nfly), /informational
                endif
                ; are there any gloves?
                if (art_camera_sample.ngl gt 0) then begin
                    message, 'Ignored any gloves (?) for sample frame ' + int2str(art_camera_sample.frame) + ', though the header indicates that there are ' + int2str(art_camera_sample.ngl), /informational
                endif
                ; create a new array of art camera data, or append to existing?
                if (n_tags(art_camera_data) eq 0) then begin
                    art_camera_data = [art_camera_sample]
                endif else begin
                    art_camera_data = [temporary(art_camera_data), art_camera_sample]
                endelse
            endif

            ; reset data for next sample
            art_camera_sample = -1
            art_camera_markers = -1
            art_camera_bodies = -1
        endif

        ; is a new art camera sample starting? 
        acs_positions = stregex(current_line, acs_pattern, length=acs_lengths, /subexpr)
        if (acs_positions[0] ge 0) then begin
            acs_fields = strmid(current_line, acs_positions[1:*], acs_lengths[1:*])
            ;message, 'An ART Camera Sample was encountered: ' + strjoin(acs_fields, ' '), /informational 

            ; is some other art camera sample is already beeing written?
            if (n_tags(art_camera_sample) gt 0) then begin
                ptr_free, art_camera_sample.markers
                ptr_free, art_camera_sample.bodies
                message, 'An ART Camera Sample was not ended by a blank line and is therefore discarded!', /informational
            endif
            ; start a new art camera sample
            art_camera_sample = {art_camera_sample, acs_fields[0], acs_fields[1], acs_fields[2], acs_fields[3], acs_fields[4], acs_fields[5], acs_fields[6], acs_fields[7], ptr_new(), ptr_new()}
            art_camera_markers = -1;
            art_camera_bodies = -1
        endif

        ; is an art camera marker given?
        acm_positions = stregex(current_line, acm_pattern, length=acm_lengths, /subexpr)
        if (acm_positions[0] ge 0) then begin
            acm_fields = strmid(current_line, acm_positions[1:*], acm_lengths[1:*])
            ;message, 'ART Camera Marker data where encountered: ' + strjoin(acm_fields, ' '), /informational

            ; was a valid art camera sample header encountered?
            if (n_tags(art_camera_sample) gt 0) then begin
                ; create a new array of art marker positions, or append to existing?
                marker = {art_camera_marker, acm_fields[0], acm_fields[1], [acm_fields[2], acm_fields[3], acm_fields[4]]}
                if (n_tags(art_camera_markers) eq 0) then begin
                    art_camera_markers = [marker]
                endif else begin
                    ; does a marker with the same id already exist in this sample?
                    ids = where(art_camera_markers.mar eq marker.mar, n_ids)
                    if (n_ids gt 0) then begin
                        message, 'An ART Camera Marker with id ' + int2str(marker.mar) + ' already exists in the current ART Camera Sample. This duplicate is therefore ignored!', /informational
                    endif else begin
                        art_camera_markers = [temporary(art_camera_markers), marker]
                    endelse
                endelse 
            endif else begin
                message, 'Ignoring ART Camera Marker data since there is no corresponding ART Camera Sample header!', /informational
            endelse
        endif
    
        ; is an art camera body given?
        acb_positions = stregex(current_line, acb_pattern, length=acb_lengths, /subexpr)
        if (acb_positions[0] ge 0) then begin
            acb_fields = strmid(current_line, acb_positions[1:*], acb_lengths[1:*])
            ;message, 'ART Camera Body data where encountered: ' + strjoin(acb_fields, ' '), /informational

            ; was a valid art camera sample header encountered?
            if (n_tags(art_camera_sample) gt 0) then begin
                ; create a new array of art body positions, or append to existing?
                body = {art_camera_body, acb_fields[0], acb_fields[1], [acb_fields[2], acb_fields[3], acb_fields[4]], [acb_fields[5], acb_fields[6], acb_fields[7]], transpose([[acb_fields[8], acb_fields[9], acb_fields[10]], [acb_fields[11], acb_fields[12], acb_fields[13]], [acb_fields[14], acb_fields[15], acb_fields[16]]])}
                if (n_tags(art_camera_bodies) eq 0) then begin
                    art_camera_bodies = [body]
                endif else begin
                    ; does a body with the same id already exist in this sample?
                    ids = where(art_camera_bodies.bod eq body.bod, n_ids)
                    if (n_ids gt 0) then begin
                        message, 'An ART Camera Body with id ' + int2str(marker.mar) + ' already exists in the current ART Camera Sample. This duplicate is therefore ignored!', /informational
                    endif else begin
                        art_camera_bodies = [temporary(art_camera_bodies), body]
                    endelse
                endelse 
            endif else begin
                message, 'Ignoring ART Camera Body data since there is no corresponding ART Camera Sample header!', /informational
            endelse
        endif

    endwhile
    free_lun, lun

    ; wrap loaded data in this object
    self.data = ptr_new(art_camera_data)

    ;profiling code
    t1 = systime(1)
    message, 'LOAD_ASCII executed in ' + string(t1 - t0) + ' seconds', /informational 

END


;; Loads ART camera data from a SAV file.
;; The file must contain data in IDL's SAV format, in a certain structure. The required structure is the same 
;; as created by the save_sav method. 
;; TODO improve checks for integrity of loaded data
;; @param FILENAME (IN) The full path to the SAV file. It can be absolute or relative.
;;
PRO ART_CAMERA_DATA::LOAD_SAV, FILENAME

    ; check if ART data have already been loaded
    if (ptr_valid(self.data)) then begin
        message, 'Cannot load ART data from file, since some data are already loaded!'
    endif

    ; load ART data from file
    restore, filename
    
    ; check ART data
    if (n_elements(art_camera_data) eq 0) then begin
        message, 'Given file does not contain any valid ART data!'
    endif

    ; wrap loaded data in this object
    self.data = ptr_new(art_camera_data)

END


;; Saves ART camera data to a SAV file.
;; The created file uses IDL's SAV format, in a certain structure. The contents of the file can be restored 
;; by the load_sav method later.
;; TODO check filename
;; @param FILENAME (IN) The full path to the SAV file. It can be absolute or relative.
;; 
PRO ART_CAMERA_DATA::SAVE_SAV, FILENAME

    ; show file picker dialog if needed
    if (n_elements(filename) eq 0) then begin
        filename = dialog_pickfile(title='Save ART Camera Data To File', filter=['*.sav'], /write, /overwrite_prompt, /fix_filter)
    endif

    ; check file
    if (filename eq '') then begin
        message, 'No valid file was specified!'
    end
        
    ; save ART data
    art_camera_data = *(self.data)
    save, art_camera_data, filename=filename

END


;; A data access method to retrieve marker locations from this ART Camera Data.
;; The ART Camera Data samples each contain an arbitary amount of markers, and one marker may 
;; have different IDs in different sample frames, if it was occluded inbetween.
;; This function tries to consolidate the situation by determining how many markers there
;; really are in the whole data sequence. Then an appropriate number of marker slots is set up, 
;; and filled with the marker data from all sample frames consecutively. If the ID of a marker already 
;; appeared in the previous sample frame, the marker is assigned to the corresponding slot.
;; Each newly appearing marker is assigned to the best fitting free slot, according to its 
;; distance from the location used in that slot most recently. A free slot occurs, when a sample frame
;; contains less markers than there are marker slots. 
;; @param DISTANCE_THRESHOLD (?IN) A distance threshold for markers to be assigned to slots after they
;;                                 where temporarily occluded. A marker is only assigned to the slot, if 
;;                                 its distance from the last marker in the slot is less than the
;;                                 distance threshold. The distance threshold is interpreted as a number 
;;                                 in the units of the ART coordinate system (usually millimeters). 
;;                                 Defaults to infinity, if omitted.
;; @return A three-dimensional array of marker data. 
;;         The first dimension corresponds to the sample frames. 
;;         The second dimension corresponds to the slots for different markers. 
;;         The third dimension corresponds to the 3D spatial coordinates of the markers in the 
;;         global ART coordinate system.
;;         A free marker slot is marked by setting all the marker coordinates to NaN.
;;
FUNCTION ART_CAMERA_DATA::GET_MARKERS, DISTANCE_THRESHOLD

    ; setup maximum (square) distance threshold from argument, or use default if not given
    if (n_elements(distance_threshold) eq 1) then begin
        max_distance = distance_threshold ^ 2
    endif else begin
        max_distance = !values.f_infinity
        message, 'No threshold for maximum marker distance was given. Assuming default: Infinity.', /informational
    endelse

    ; get the maximum number of markers
    data = *(self.data)
    min_markers = -1
    max_markers = -1
    for i = 0ul, n_elements(data)-1 do begin
        if (ptr_valid(data[i].markers)) then begin
            n_markers = n_elements(*(data[i].markers))
        endif else begin
            n_markers = 0 
        endelse
        if (max_markers lt n_markers) then max_markers = n_markers
        if ((min_markers gt n_markers) or (min_markers lt 0)) then min_markers = n_markers
    endfor
    message, 'Minimum number of markers: ' + int2str(min_markers), /informational
    message, 'Maximum number of markers: ' + int2str(max_markers), /informational

    ; create an array of approprite size for all marker data
    ; (1st dimension = frames, 2nd dimension = slots for markers, 3rd dimension = space dimensions)
    art_marker_array = rebin([!values.f_nan], n_elements(data), max_markers, 3)

    ; create an array to hold id and location of the last valid marker for each slot
    last_marker_ids = rebin([-1], max_markers)
    last_marker_locations = rebin([!values.f_nan], 3, max_markers)

    ; fill all frames of the array with marker data
    for i = 0ul, n_elements(data)-1 do begin

        ; any valid markers for current frame?
        if (ptr_valid(data[i].markers)) then begin

            ; get markers for current frame
            markers = *(data[i].markers)
            n_markers = n_elements(markers)

            ; find all pairs of slots and markers which have corresponding ids
            id_matrix = rebin([last_marker_ids], max_markers, n_markers) eq transpose(rebin([markers.mar], n_markers, max_markers))
            ids = where(id_matrix, n_ids)
            if (n_ids gt 0) then begin
                ; get mapping of markers to slots
                slot_indices = ids mod max_markers
                marker_indices = ids / max_markers

                ; fill all slots with their corresponding markers by id
                art_marker_array[i,slot_indices,*] = (transpose(reform(markers[marker_indices].loc, 3, n_ids, 1, /overwrite), [2,1,0]))
                last_marker_locations[*,slot_indices] = markers[marker_indices].loc
                last_marker_ids[slot_indices] = markers[marker_indices].mar

                ; could all markers be assigned by id? 
                if (n_elements(marker_indices) eq n_markers) then begin
                    ; move on to next frame
                    continue
                endif

                ; get all slots and markers that where not assigned yet
                slot_indices2 = indgen(max_markers)
                slot_indices2[slot_indices] = -1
                slot_indices2 = slot_indices2[where(slot_indices2 ge 0)]
                n_slot_indices2 = n_elements(slot_indices2)
                marker_indices2 = indgen(n_markers)
                marker_indices2[marker_indices] = -1
                marker_indices2 = marker_indices2[where(marker_indices2 ge 0)]
                n_marker_indices2 = n_elements(marker_indices2)
                
            endif else begin

                ; get all slots and markers, since no could be assigned yet
                slot_indices2 = indgen(max_markers)
                n_slot_indices2 = max_markers
                marker_indices2 = indgen(n_markers)
                n_marker_indices2 = n_markers

            endelse

            ; calculate (square-)distances of all pairs of unasigned slots and markers
            distance_matrix = rebin(last_marker_locations[*,slot_indices2], 3, n_slot_indices2, n_marker_indices2)
            distance_matrix -= transpose(reform(rebin((markers.loc)[*,marker_indices2], 3, n_marker_indices2, n_slot_indices2), 3, n_marker_indices2, n_slot_indices2, /overwrite), [0,2,1])
            distance_matrix = total(distance_matrix ^ 2, 1)

            ; assign markers to slots by shortest distances
            ; if no distance could be calculated (NaN), markers and slots are assigned arbitarily
            for j = 0ul, n_marker_indices2-1 do begin

                ; search for the slot with the shortest distance to a marker
                min_distance = min(distance_matrix, min_index, /nan)

                ; if no sufficiently close slot exists, search for any free slot
                if ((min_distance ge max_distance) or (finite(min_distance, /nan))) then begin
                    min_index = (where(finite(distance_matrix, /nan)))[0]

                    ; if no slots are free due to the maximum distance threshold, introduce new ones
                    if (min_index lt 0) then begin

                        ; get indices of all markers, which are still not asigned to slots
                        unassigned_marker_indices = where(total(finite(distance_matrix), 1), n_unassigned_markers)
                        unassigned_markers = marker_indices2[unassigned_marker_indices]

                        ; create slots for new markers
                        new_art_marker_array = rebin([!values.f_nan], n_elements(data), max_markers+n_unassigned_markers, 3)
                        new_art_marker_array[*,0:max_markers-1,*] = art_marker_array
                        art_marker_array = new_art_marker_array

                        ; assign markers to new slots
                        ; TODO check if array dimensions and indices are handled correctly during assingment
                        art_marker_array[i,max_markers:*,*] = transpose(markers[unassigned_markers].loc)
                        last_marker_locations = [[last_marker_locations], [markers[unassigned_markers].loc]]
                        last_marker_ids = [last_marker_ids, markers[unassigned_markers].mar]

                        ; increase number of slots and continue to next frame
                        max_markers += n_unassigned_markers
                        message, 'Maximum number of markers was resized: ' + int2str(max_markers), /informational
                        break
                    endif
                endif
                min_slot = slot_indices2[min_index mod n_slot_indices2]
                min_marker = marker_indices2[min_index / n_slot_indices2]

                ; assign marker to slot
                art_marker_array[i,min_slot,*] = markers[min_marker].loc
                last_marker_locations[*,min_slot] = markers[min_marker].loc
                last_marker_ids[min_slot] = markers[min_marker].mar
                
                ; invalidate the corresponding row and column in the distance matrix by setting it to infinite
                distance_matrix[min_index mod n_slot_indices2, *] = !values.f_infinity
                distance_matrix[*,min_index / n_slot_indices2] = !values.f_infinity
            endfor

        endif
    endfor

    return, art_marker_array

END


;; A data access method to retrieve ART body locations from this ART Camera Data.
;; The ART data samples each contain an arbitary amount of bodies, each identified by
;; a unique ID. Such an ID is specified by the ART user during body calibration, and is
;; preserved even when the body is not visible inbetween. However, when the body is not visible 
;; during a certain period, the corresponding sample frames do not contain any data at all related 
;; to a its ID.
;; @param ID (IN) The numerical ID of the ART body whose data to get.
;; @return A three-dimensional array of body data. 
;;         The first dimension corresponds to the sample frames. 
;;         The second and third dimensions correspond to a 4x4 homogenous matrix describing
;;         the transformation from the global ART coordinate to the local coordinates of
;;         the body.
;;         Sample frames not containing any data for the body are indicated by a transformation matrix
;;         filled with NaN values.
;;
FUNCTION ART_CAMERA_DATA::GET_BODY, ID

    ; check body id parameter
    if (n_elements(id) eq 0) then begin
        message, 'No body ID given!'
    endif

    ;; create an array of NaNs with approprite size for all marker data
    ;; (1st dimension = frames, 2nd & 3rd dimension = homogenous matrix describing body transformation)
    data = *(self.data)
    art_body_array = rebin([!values.f_nan], n_elements(data), 4, 4, /sample)
    ;; fill the array with body data
    for i = 0ul, n_elements(data)-1 do begin
        ; are there any bodies in current sample?
        if (ptr_valid(data[i].bodies)) then begin
            bodies = *(data[i].bodies)
            ; is the body with the given id in current sample?
            ids = where(bodies.bod eq id, n_ids)
            if (n_ids eq 1) then begin
                body = bodies[ids]
                ; setup the homogenous transformation matrix for the body
                art_body_array[i,0:2,0:2] = body.rot
                art_body_array[i,3,0:2] = body.loc
                art_body_array[i,0:2,3] = 0.0
                art_body_array[i,3,3] = 1.0
            endif
        endif 
    endfor

    return, art_body_array

END


;; A data access method to retrieve the sampling time of this ART Camera Data.
;; That is, the average time between two consecutive ART sample frames.
;; This sampling time is derived from the time stamps stored in each sample frame. Since
;; ART Camera Data are usually sampled regularily, the average sampling time approximates the actual
;; sampling time between two arbitary consecutive frames quite well. 
;; (Note that the timestamps are usually reset at midnight, and this method will hence not 
;; calculate the correct sampling time for such ART Camera Data.)
;; @key TIMES (?OUT) An optional output argument to store an array of the actual times between all pairs of samples.
;;                   The mean of the array will always correspond to the sampling time itself. However, if the 
;;                   variance of the array is big, the average sampling time is not a good aproximation for the 
;;                   actual sampling times.
;; @return The average sampling time.
;;
FUNCTION ART_CAMERA_DATA::GET_SAMPLING_TIME, TIMES=TIMES
    
    ; check if there are enough samples to determine a sampling rate
    data = *(self.data)
    n_samples = n_elements(data)
    if n_samples lt 2 then begin
        message, 'There must be at least two samples to determine sampling time!'
    endif

    ; calculate sampling time
    duration = data[n_samples-1].ts - data[0].ts
    sampling_time = duration / double(n_samples-1)

    ; optionally calculate an array of all the actual sampling times
    if arg_present(times) then begin
        times = data[indgen(n_samples-1) + 1].ts - data.ts
    endif

    return, sampling_time

END


;; A data access method to retrieve the average Nyquist-Frequency of this ART Camera Data.
;; The Nyquist-Frequency in Hz is defined as the reciprocal of two times the sampling time in s.
;; @return The Nyquist-Frequency.
;;
FUNCTION ART_CAMERA_DATA::GET_NYQUIST_FREQUENCY

    return, 0.5 / self->get_sampling_time()

END


;; A data access method to retrieve the overall count of samples comprising this ART Camera Data.
;; @return The sample count.
;;
FUNCTION ART_CAMERA_DATA::GET_SAMPLE_COUNT

    data = *(self.data)
    return, n_elements(data)

END


;; A data access method to retrieve the overall duration of this sequence of ART Camera Data.
;; @return The duration given in seconds.
;;
FUNCTION ART_CAMERA_DATA::GET_DURATION

    data = *(self.data)
    n_samples = n_elements(data)
    return, data[n_samples-1].ts - data[0].ts

END
