;; This implementation of ART Processor is capable of 
;; retrieving marker data. It can also interpolate missing marker 
;; samples and can be setup to filter for relevant markers.
;; @field DISTANCE_THRESHOLD The distance threshold to pass to the get_markers method of the ART Camera Data.
;; @field RELEVANT_MARKERS A pointer to an array containing all markers considered relevant for processing.
;;
PRO ART_MARKER_PROCESSOR__DEFINE

    ; define the class structure 
    struct = { ART_MARKER_PROCESSOR, INHERITS art_processor, distance_threshold:0.0, relevant_markers:ptr_new() }

END


;; Creates a new ART Marker Processor. 
;; The ART Marker Processor is optionally bound to the ART Camera Data given. 
;; The distance_threshold and the list of relevant markers can also be predefined optionally.
;; @param DATA (?IN) The ART Camera Data to setup initially. (See superclass.)
;; @key DISTANCE_THRESHOLD (?IN) The distance_threshold to setup initially. (See method set_distance_threshold.)
;; @key RELEVANT_MARKERS (?IN) The list of relevant markers to setup initially. (See method set_relevant_markers.)
;;
FUNCTION ART_MARKER_PROCESSOR::INIT, DATA, DISTANCE_THRESHOLD=DISTANCE_THRESHOLD, RELEVANT_MARKERS=RELEVANT_MARKERS

    ; initialize superclass
    if (self->art_processor::init(data) eq 0) then begin
        return, 0
    endif 

    ; try to initialize field values
    catch, e
    if (e ne 0) then begin

        message, !error_state.msg, /informational
        return, 0

    endif

        ; try to initialize field values
        self->set_distance_threshold, distance_threshold
        self->set_relevant_markers, relevant_markers

    catch, /cancel
    return, 1

END


;; Cleans up the heap upon destruction.
;; More specifically, the pointer used internally to store the
;; list of relevant markers is freed. 
;; Also, the cleanup method of the superclass is invoked.
;;
PRO ART_MARKER_PROCESSOR::CLEANUP

    self->art_processor::cleanup

    ; clean up list of relevant markers
    if (ptr_valid(self.relevant_markers)) then begin
        heap_free, self.relevant_markers
    endif

END


;; Sets up the distance threshold to use when retrieving raw marker data.
;; This value is passed to the get_markers method of the ART Camera Data upon processing.
;; If no value is given, the distance threshold is reset to infinity.
;; TODO Check that the given distance is a positive number.
;; @param DISTANCE_THRESHOLD (?IN) The distance threshold to setup this ART Marker Processor for.
;;                                 This must be a numeric value.
;;
PRO ART_MARKER_PROCESSOR::SET_DISTANCE_THRESHOLD, DISTANCE_THRESHOLD

    ; reset to infinity, if no value given
    if (n_elements(distance_threshold) eq 0) then begin
        self.distance_threshold = !values.f_infinity
        return
    endif

    self.distance_threshold = distance_threshold

END


;; Returns the distance threshold setup for this ART Marker Processor. 
;; @return The distance threshold.
;;
FUNCTION ART_MARKER_PROCESSOR::GET_DISTANCE_THRESHOLD

    return, self.distance_threshold

END


;; Sets up a list of markers relevant for further processing. 
;; When retrieving marker data all markers, whose index is not in the given list, are discarded.
;; Also see the method find_relevant_markers, which can setup such a list semi-automatically.
;; TODO Check that an array of integer numbers is given, that it contains no duplicates and is that it is ordered. (Or remove duplicates and perform ordering.)
;; @param RELEVANT_MARKERS (?IN) The list of relevant markers to setup this ART Marker Processor for.
;;                               This must be an array of integers ordered ascendingly and not containing any duplicates.
;;                               Ommit this paramter or set it to -1 to reset the list of relevant markers.
;;
PRO ART_MARKER_PROCESSOR::SET_RELEVANT_MARKERS, RELEVANT_MARKERS

    ; clean up old list of relevant markers
    if (ptr_valid(self.relevant_markers)) then begin
        heap_free, self.relevant_markers
    endif

    ; reset to null, if no value given
    if (n_elements(relevant_markers) eq 0) then begin
        self.relevant_markers = ptr_new()
        return
    endif
    ; reset to null, if explicitly requested
    if (n_elements(relevant_markers) eq 1) then begin
        if (relevant_markers eq -1) then begin
            self.relevant_markers = ptr_new()
            return
        endif
    endif

    self.relevant_markers = ptr_new(relevant_markers)

END


;; Returns the list of relevant markers setup for this ART Marker Processor. 
;; @return The list of relevant markers.
;;
FUNCTION ART_MARKER_PROCESSOR::GET_RELEVANT_MARKERS

    ; does the list of relevant markers contain any entries?
    if (ptr_valid(self.relevant_markers)) then begin
        return, *(self.relevant_markers)
    endif else begin
        return, -1
    endelse

END


;; Semi-automatically sets up the list of relevant markers according to given criteria.
;; The creteria supported are related to the amount of valid sample frames available for the 
;; respective marker. A minimum value can either given an absolute count of sample frames or as a fraction 
;; of all sample frames in the given sequence. 
;; @key MIN_SAMPLE_COUNT (?IN) The absolute count of valid sample frames a marker must have to be considered relevant. 
;;                             Must be an integer number. Defaults to 0.
;; @key MIN_SAMPLE_FRACTION (?IN) The fraction of valid sample frames a marker must have to be considered relevant.
;;                                Must be a numeric value. Defaults to 0.0.
;;
PRO ART_MARKER_PROCESSOR::FIND_RELEVANT_MARKERS, MIN_SAMPLE_COUNT=MIN_SAMPLE_COUNT, MIN_SAMPLE_FRACTION=MIN_SAMPLE_FRACTION, MIN_SAMPLE_DURATION=MIN_SAMPLE_DURATION

    ; is this processor setup properly?
    if (self->is_processible() eq 0) then begin
        message, 'Setup all fields of this ART_MARKER_PROCESSOR before using it!'
    endif

    ; check filter arguments an use the more restrictive one
    if ((n_elements(min_sample_count) eq 0) and (n_elements(min_sample_fraction) eq 0) and (n_elements(min_sample_duration) eq 0)) then begin
        message, 'At least one of the arguments MIN_SAMPLE_COUNT or MIN_SAMPLE_FRACTION or MIN_SAMPLE_DURATION must be given!'
    endif 
    if (n_elements(min_sample_count) eq 0) then begin
        min_sample_count = 0
    endif
    if (n_elements(min_sample_fraction) eq 0) then begin
        min_sample_fraction = 0.0
    endif
    if (n_elements(min_sample_duration) eq 0) then begin
        min_sample_duration = 0.0
    endif

    ; use the method of the superclass to get raw markers
    markers = (self->get_art_data())->get_markers(self.distance_threshold)

    ; find the most strict one of the count, fraction and duration requirements
    sample_count = (self->get_art_data())->get_sample_count()
    sampling_time = (self->get_art_data())->get_sampling_time()
    min = min_sample_count > (min_sample_fraction * sample_count) > (min_sample_duration / sampling_time)

    ; find missing samples for each marker, and compare them to the requirements
    valid_marker_samples = finite(transpose(total(markers, 3)))
    sample_counts = total(valid_marker_samples, 2)
    relevant_markers = where(sample_counts ge min)

    ; setup the relevasnt markers accordingly
    self->set_relevant_markers, relevant_markers

END


;;
;;
FUNCTION ART_MARKER_PROCESSOR::RETRIEVE_MARKER_COUNT

    ; is this processor setup properly?
    if (self->is_processible() eq 0) then begin
        message, 'Setup all fields of this ART_MARKER_PROCESSOR before using it!'
    endif
    
    ; get raw marker data
    markers = (self->get_art_data())->get_markers(self.distance_threshold)
    marker_dimensions = size(markers, /dimensions)

    ; return marker count
    return, marker_dimensions[1]

END


;; A processing method for retrieving a sequence of ART marker locations.
;; This method first gets all marker data from the ART Camera Data setup. 
;; It then filters the marker data for the markers setup as relevant. 
;; Finally it interpolates missing samples for those markers, and returns the result.
;; @return A three-dimensional array of ART marker data.
;;         The first dimension corresponds to the sample frames. 
;;         The second dimension corresponds to the slots of relevant markers. 
;;         The third dimension corresponds to the 3D spatial coordinates of the markers.
;;
FUNCTION ART_MARKER_PROCESSOR::RETRIEVE_MARKERS

    ; is this processor setup properly?
    if (self->is_processible() eq 0) then begin
        message, 'Setup all fields of this ART_MARKER_PROCESSOR before using it!'
    endif
    
    ; get raw marker data
    markers = (self->get_art_data())->get_markers(self.distance_threshold)

    ; discard irrelevant markers as explicitly specified
    if (ptr_valid(self.relevant_markers)) then begin
        markers = markers[*,reform([*(self.relevant_markers)]),*]
    endif

    ; interpolate any missing samples of markers
    self->interpolate_missing_markers, markers

    ; return processed markers
    return, markers
    
END


;; This internal method interpolates missing samples in a sequence of ART marker locations. 
;; Each marker sample containing NaNs is interpolated using valid data from the
;; same marker in nearby sample frames. If a marker is missing inbetween two 
;; valid samples, the data is interpolated lineary over the missing range. If a 
;; marker is missing in the beginning or a the end of the sequence, the data of the
;; next valid sample is just repeated.
;; @param MARKERS (IN/OUT) A three-dimensional array of ART marker data, as returned by the get_markers method of ART Camera Data.
;; 
PRO ART_MARKER_PROCESSOR::INTERPOLATE_MISSING_MARKERS, MARKERS

    ; check marker array
    if (size(markers, /n_dimensions) ne 3) then begin
        message, 'The array of markers must have exactly three dimensions! (Sample frames, markers, 3D-coordinates)'
    endif
    marker_dimensions = size(markers, /dimensions)
    if (marker_dimensions[2] ne 3) then begin
        message, 'The array of markers must contain 3d vectors!'
    endif

    ; find any missing marker samples
    n_markers = marker_dimensions[1]
    n_frames = marker_dimensions[0]
    missing = where(finite(transpose(total(markers, 3)), /nan), n_missing)

    ; do any marker samples have to be interpolated?
    if (n_missing gt 0) then begin

        ; calculate marker ids and frame numbers of all missing marker samples
        missing_markers = missing mod n_markers
        missing_frames = missing / n_markers

        ; check all marker slots for missing samples
        for mar = 0ul, n_markers-1 do begin

            ; get indices of all frames where the current marker is missing
            frame_indices = where(missing_markers eq mar, n_missing)

            ; is the current marker missing in any frame?
            if (n_missing gt 0) then begin

                ; get frame numbers of all frames where the current marker is missing
                frames = missing_frames[frame_indices]

                ; get frame numbers adjunctant to frames where the current marker is missing
                lower_margin_indices = where(frames ne [-1, frames+1])
                lower_margins = frames[lower_margin_indices] - 1
                upper_margin_indices = where(frames ne [frames[1:*]-1, -1])
                upper_margins = frames[upper_margin_indices] + 1

                ; assert that everything is still ok
                if (n_elements(lower_margins) ne n_elements(upper_margins)) then begin
                    message, 'Internal error: Calculated numbers of lower and upper margins are not equal!'
                endif

                ; iterate over each range of subsequent frames where the current marker is missing
                for range = 0ul, n_elements(lower_margins)-1 do begin
                    lower_margin = lower_margins[range]
                    upper_margin = upper_margins[range]

                    ; is there any data at all to interpolate or extrapolate from for current range?
                    if ((lower_margin lt 0) and (upper_margin ge n_frames)) then begin
                        ; reset missing data to 0 to avoid problems with NaN in later calculations
                        markers[*,mar,*] = 0.0
                        message, 'No valid data to interpolate marker ' + int2str(mar) + ' from. Resetting it to null vectors!' , /informational
                        continue
                    endif

                    ; is the current range the beginning?
                    if (lower_margin lt 0) then begin
                        ; extrapolate the beginning range
                        ending_location = reform(markers[upper_margin,mar,*])
                        markers[0:upper_margin,mar,*] = rebin(reform(ending_location, 1, 1, 3), upper_margin+1, 1, 3)
                        continue
                    endif

                    ; is the current range the ending?
                    if (upper_margin ge n_frames) then begin
                        ; extrapolate the ending range
                        beginning_location = reform(markers[lower_margin,mar,*])
                        markers[lower_margin:n_frames-1,mar,*] = rebin(reform(beginning_location, 1, 1, 3), n_frames-lower_margin, 1, 3)
                        continue
                    endif

                    ; interpolate range from both sides 
                    range_length = upper_margin-lower_margin+1
                    beginning_location = reform(markers[lower_margin,mar,*])
                    ending_location = reform(markers[upper_margin,mar,*])
                    interpolated_range = congrid([[beginning_location], [ending_location]], 3, range_length, /interp, /minus_one)
                    markers[lower_margin:upper_margin,mar,*] = transpose(reform(interpolated_range, 3, range_length, 1), [1,2,0])
                endfor
            end

        endfor

    endif

END
