;; This is an abstract superclass for any class capable of 
;; processing ART Camera Data. This class just offers methods to setup the
;; ART Camera Data to process. Implementing classes should offer methods to setup 
;; other processing parameters, too. Most importantly, they should offer methods
;; performing different kinds of processing on the ART Camera Data and returning the 
;; results. Examples of the processing applied would be interpolation, 
;; coordinate-system-transformations, noise-filtering and other data analysis. 
;; The processing methods themselves are not predefined by this class, but rather
;; left to implementing classes.
;; @field DATA A pointer to the ART Camera Data to process.
;;
PRO ART_PROCESSOR__DEFINE

    ; define the class structure 
    struct = { ART_PROCESSOR, data:obj_new() }

END


;; Creates a new ART Processor. 
;; The ART Processor is optionally bound to the ART Camera Data given. 
;; @param DATA (?IN) The ART Camera Data to setup initially. (See method set_art_data.)
;;
FUNCTION ART_PROCESSOR::INIT, DATA

    ; try to set initial ART data
    catch, e
    if (e ne 0) then begin

        message, !error_state.msg, /informational
        return, 0

    endif

        ; try to set initial ART data
        self->set_art_data, data
        return, 1

    catch, /cancel

END


;; Cleans up upon destruction. 
;; This ART Processor does not need to perform any clean up.
;;
PRO ART_PROCESSOR::CLEANUP
END


;; Sets up the ART Camera Data to be processed by this ART Processor. 
;; If no ART Camera Data are given, any ART Camera Data previously setup for this 
;; ART Processor are reset.
;; @param DATA (?IN) The ART Camera Data to setup this ART Processor for.
;;
PRO ART_PROCESSOR::SET_ART_DATA, DATA

    ; is any parameter given?
    if (n_elements(data) eq 0) then begin

        ; reset ART data
        self.data = obj_new()
        return

    endif

    ; are valid ART camera data given?
    if (obj_valid(data)) then begin
        if (obj_isa(data, 'art_camera_data')) then begin

            ; set ART data
            self.data = data
            return

        endif
    endif

    ; no valid data are given
    message, 'The given data must be of type ART_CAMERA_DATA!'

END


;; Returns the ART Camera Data setup for this ART Processor. 
;; If no ART Camera Data are currently setup, a null-object is returned.
;; @return The ART Camera Data.
;;
FUNCTION ART_PROCESSOR::GET_ART_DATA

    return, self.data

END


;; Checks if this ART Processor is setup correctly. 
;; That is, if it is in a state allowing for usefull processing.
;; This implementation of the method just checks, if any valid
;; ART Camera Data are setup. Subclasses should overwrite this method, 
;; to check all other parameters needed for processing, too.
;; The subclasses' processing methods should check the result of
;; this method as precondition. They should issue an error, if the 
;; result is 0.
;; @return 1, if this ART Processor is setup correctly, 0 otherwise.
;;
FUNCTION ART_PROCESSOR::IS_PROCESSIBLE

    ; are the ART camera data setup?
    if (obj_valid(self.data) eq 0) then begin
        message, 'No ART camera data are setup currently.', /informational
        return, 0
    end

    return, 1

END
