;; This class is a cache-enabled version of its superclass ART Camera Data.
;; It stores the ART Camera Data internally and offers the same 
;; data access methods as the superclass. However, the data access methods perform 
;; some caching, so that data are retrieved much faster, if they have already been retrieved 
;; before. However, this caching consumes additional memory.
;; @field CACHED_MARKERS A pointer to cache the last marker data retrieved. Or null if none retrieved yet.
;; @field CACHED_MARKERS_DISTANCE_THRESHOLD The distance threshold, for which the cached marker data were retrieved.
;; @field CACHED_BODY A pointer to cache the data of the last body retrieved. Or null if none retrieved yet.
;; @field CACHED_BODY_ID The id of the of the cached body data.
;; 
PRO CACHED_ART_CAMERA_DATA__DEFINE

    ; define the class structure 
    struct = { CACHED_ART_CAMERA_DATA, INHERITS art_camera_data, cached_markers:ptr_new(), cached_markers_distance_threshold:0.0, cached_body:ptr_new(), cached_body_id:0 }

END


;; Does the same as the corresponding constructor of the superclass.
;;
FUNCTION CACHED_ART_CAMERA_DATA::INIT, FILENAME, ASCII=ASCII, SAV=SAV

    return, self->art_camera_data::init(filename, ascii=ascii, sav=sav)

END


;; Cleans up the heap upon destruction.
;; More specifically the superclass is cleaned up, and 
;; all cached data are freed from the heap.
;;
PRO CACHED_ART_CAMERA_DATA::CLEANUP

    self->art_camera_data::cleanup

    if ptr_valid(self.cached_markers) then begin
        heap_free, self.cached_markers
    endif

    if ptr_valid(self.cached_body) then begin
        heap_free, self.cached_body
    endif

END


;; Retrieves the marker data, the same way the corresponding method of the superclass does.
;; Additionally some caching is applied. That is, when the marker data has been retrieved for a given
;; distance threshold, it is cached. The next requests for markers with the same distance threshold 
;; does not need to be recalculated, but is drawn from the cache.
;;
FUNCTION CACHED_ART_CAMERA_DATA::GET_MARKERS, DISTANCE_THRESHOLD

    ; check distance theshold parameter
    if (n_elements(distance_threshold) eq 0) then begin
        distance_threshold = !values.f_infinity
        message, 'No threshold for maximum marker distance was given. Assuming default: Infinity.', /informational
    endif

    ; invalidate marker cache, if necessary
    if ((ptr_valid(self.cached_markers)) and (self.cached_markers_distance_threshold ne distance_threshold)) then begin
        heap_free, self.cached_markers
    endif

    ; update marker cache, if necessary
    if (~ptr_valid(self.cached_markers)) then begin
        markers = self->art_camera_data::get_markers(distance_threshold)
        self.cached_markers = ptr_new(markers)
        self.cached_markers_distance_threshold = distance_threshold
    endif
    
    ; return cached markers
    return, *(self.cached_markers)

END


;; Retrieves the body data, the same way the corresponding method of the superclass does.
;; Additionally some caching is applied. That is, when a body with a given id is retrieved,
;; it is cached. The next requests for markers with the same id does not need to be 
;; recalculated, but is drawn from the cache.
;; TODO Implement caching for more than one body id? A map of known body ids could be used...
;;
FUNCTION CACHED_ART_CAMERA_DATA::GET_BODY, ID

    ; check body id parameter
    if (n_elements(id) eq 0) then begin
        message, 'No body ID given!'
    endif

    ; invalidate body cache, if necessary
    if ((ptr_valid(self.cached_body)) and (self.cached_body_id ne id)) then begin
        heap_free, self.cached_body
    endif

    ; update body cache, if necessary
    if (~ptr_valid(self.cached_body)) then begin
        body = self->art_camera_data::get_body(id)
        self.cached_body = ptr_new(body)
        self.cached_body_id = id
    endif

    ; return cached body
    return, *(self.cached_body)

END
